(* Guestfs Browser.
 * Copyright (C) 2010 Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *)

open ExtString
open ExtList
open Unix
open Printf

open Utils
open DeviceSet

open Filetree_type
open Filetree_ops

module G = Guestfs

type t = Filetree_type.t

let rec create ~packing () =
  let view = GTree.view ~packing () in
  (*view#set_rules_hint true;*)
  (*view#selection#set_mode `MULTIPLE; -- add this later *)

  (* Hash of index numbers -> hdata.  We do this because it's more
   * efficient for the GC compared to storing OCaml objects directly in
   * the rows.
   *)
  let hash = Hashtbl.create 1023 in

  (* The columns stored in each row.  The hidden [index_col] column is
   * an index into the hash table that records everything else about
   * this row (see hdata above).  The other display columns, eg.
   * [name_col] contain Pango markup and thus have to be escaped.
   *)
  let cols = new GTree.column_list in
  (* Hidden: *)
  let index_col = cols#add Gobject.Data.int in
  (* Displayed: *)
  let mode_col = cols#add Gobject.Data.string in
  let name_col = cols#add Gobject.Data.string in
  let size_col = cols#add Gobject.Data.string in
  let date_col = cols#add Gobject.Data.string in

  (* Create the model. *)
  let model = GTree.tree_store cols in

  (* Create the view. *)
  view#set_model (Some (model :> GTree.model));

  let renderer = GTree.cell_renderer_text [], ["markup", mode_col] in
  let mode_view = GTree.view_column ~title:"Permissions" ~renderer () in
  mode_view#set_resizable true;
  ignore (view#append_column mode_view);

  let renderer = GTree.cell_renderer_text [], ["markup", name_col] in
  let name_view = GTree.view_column ~title:"Filename" ~renderer () in
  name_view#set_resizable true;
  ignore (view#append_column name_view);

  let renderer = GTree.cell_renderer_text [`XALIGN 1.], ["markup", size_col] in
  let size_view = GTree.view_column ~title:"Size" ~renderer () in
  size_view#set_resizable true;
  ignore (view#append_column size_view);

  let renderer = GTree.cell_renderer_text [`XALIGN 1.], ["markup", date_col] in
  let date_view = GTree.view_column ~title:"Date" ~renderer () in
  date_view#set_resizable true;
  ignore (view#append_column date_view);

  let t = {
    view = view; model = model; hash = hash;
    index_col = index_col;
    mode_col = mode_col; name_col = name_col; size_col = size_col;
    date_col = date_col;
  } in

  (* Open a context menu when a button is pressed. *)
  ignore (view#event#connect#button_press ~callback:(button_press t));

  t

(* Handle mouse button press on the selected row.  This opens the
 * pop-up context menu.
 * http://scentric.net/tutorial/sec-selections-context-menus.html
 *)
and button_press ({ model = model; view = view } as t) ev =
  let button = GdkEvent.Button.button ev in
  let x = int_of_float (GdkEvent.Button.x ev) in
  let y = int_of_float (GdkEvent.Button.y ev) in
  let time = GdkEvent.Button.time ev in

  (* Right button for opening the context menu. *)
  if button = 3 then (
(*
    (* If no row is selected, select the row under the mouse. *)
    let paths =
      let sel = view#selection in
      if sel#count_selected_rows < 1 then (
        match view#get_path_at_pos ~x ~y with
        | None -> []
        | Some (path, _, _, _) ->
            sel#unselect_all ();
            sel#select_path path;
            [path]
      ) else
        sel#get_selected_rows (* actually returns paths *) in
*)
    (* Select the row under the mouse. *)
    let paths =
      let sel = view#selection in
      match view#get_path_at_pos ~x ~y with
      | None -> []
      | Some (path, _, _, _) ->
          sel#unselect_all ();
          sel#select_path path;
          [path] in

    (* Get the hdata for all the paths.  Filter out rows that it doesn't
     * make sense to select.
     *)
    let paths =
      List.filter_map (
        fun path ->
          let row = model#get_iter path in
          let hdata = get_hdata t row in
          match hdata with
          | { content=(Loading | ErrorMessage _ | Info _) } -> None
          | { content=(Top _ | Directory _ | File _) } -> Some (path, hdata)
      ) paths in

    (* Based on number of selected rows and what is selected, construct
     * the context menu.
     *)
    if paths <> [] then (
      let menu = make_context_menu t paths in
      menu#popup ~button ~time
    );

    (* Return true so no other handler will run. *)
    true
  )
  (* We didn't handle this, defer to other handlers. *)
  else false

and make_context_menu t paths =
  let menu = GMenu.menu () in
  let factory = new GMenu.factory menu in

  let item = factory#add_item "Open" in
  item#misc#set_sensitive false;

  let rec add_file_items path =
    let item = factory#add_item "File information" in
    item#misc#set_sensitive false;
    let item = factory#add_item "Checksum" in
    item#misc#set_sensitive false;
    ignore (factory#add_separator ());
    let item = factory#add_item "Download ..." in
    ignore (item#connect#activate ~callback:(download_file t path));

  and add_directory_items path =
    let item = factory#add_item "Directory information" in
    item#misc#set_sensitive false;
    let item = factory#add_item "Calculate disk usage" in
    ignore (item#connect#activate ~callback:(disk_usage t path));
    ignore (factory#add_separator ());
    let item = factory#add_item "Download ..." in
    item#misc#set_sensitive false;
    let item = factory#add_item "Download as .tar ..." in
    ignore (item#connect#activate
              ~callback:(download_dir_tarball t Slave.Tar path));
    let item = factory#add_item "Download as .tar.gz ..." in
    ignore (item#connect#activate
              ~callback:(download_dir_tarball t Slave.TGZ path));
    let item = factory#add_item "Download as .tar.xz ..." in
    ignore (item#connect#activate
              ~callback:(download_dir_tarball t Slave.TXZ path));
    let item = factory#add_item "Download list of filenames ..." in
    ignore (item#connect#activate ~callback:(download_dir_find0 t path));

  and add_os_items path =
    let item = factory#add_item "Operating system information" in
    ignore (item#connect#activate ~callback:(display_inspection_data t path));
    ignore (factory#add_separator ());
    add_volume_items path

  and add_volume_items path =
    let item = factory#add_item "Filesystem used & free" in
    item#misc#set_sensitive false;
    let item = factory#add_item "Block device information" in
    item#misc#set_sensitive false;
    ignore (factory#add_separator ());
    add_directory_items path
  in

  (match paths with
   (* single selection *)
   | [path, { content=Top (Slave.OS os)} ] ->  (* top level operating system *)
       add_os_items path

   | [path, { content=Top (Slave.Volume dev) }] -> (* top level volume *)
       add_volume_items path

   | [path, { content=Directory direntry }] ->     (* directory *)
       add_directory_items path

   | [path, { content=File direntry }] ->          (* file *)
       add_file_items path

   | [_, { content=Loading }]
   | [_, { content=ErrorMessage _ }] -> ()

   | _ ->
       (* At the moment multiple selection is disabled.  When/if we
        * enable it we should do something intelligent here. XXX
        *)
       ()
  );

  menu

(* Mark up mode. *)
let markup_of_mode mode =
  let c =
    if is_socket mode then 's'
    else if is_symlink mode then 'l'
    else if is_regular_file mode then '-'
    else if is_block mode then 'b'
    else if is_directory mode then 'd'
    else if is_char mode then 'c'
    else if is_fifo mode then 'p' else '?' in
  let ru = if is_ru mode then 'r' else '-' in
  let wu = if is_wu mode then 'w' else '-' in
  let xu = if is_xu mode then 'x' else '-' in
  let rg = if is_rg mode then 'r' else '-' in
  let wg = if is_wg mode then 'w' else '-' in
  let xg = if is_xg mode then 'x' else '-' in
  let ro = if is_ro mode then 'r' else '-' in
  let wo = if is_wo mode then 'w' else '-' in
  let xo = if is_xo mode then 'x' else '-' in
  let str = sprintf "%c%c%c%c%c%c%c%c%c%c" c ru wu xu rg wg xg ro wo xo in

  let suid = is_suid mode in
  let sgid = is_sgid mode in
  let svtx = is_svtx mode in
  if suid then str.[3] <- 's';
  if sgid then str.[6] <- 's';
  if svtx then str.[9] <- 't';

  "<span color=\"#222222\" size=\"small\">" ^ str ^ "</span>"

(* Mark up dates. *)
let markup_of_date t =
  (* Guestfs gives us int64's, we want float which is OCaml's
   * equivalent of time_t.
   *)
  let t = Int64.to_float t in

  let show_full_date () =
    let tm = localtime t in
    sprintf "<span color=\"#222222\" size=\"small\">%04d-%02d-%02d %02d:%02d:%02d</span>"
      (tm.tm_year + 1900) (tm.tm_mon + 1) tm.tm_mday
      tm.tm_hour tm.tm_min tm.tm_sec
  in

  (* How long ago? *)
  let now = time () in
  let ago = now -. t in
  if ago < 0. then (* future *)
    show_full_date ()
  else if ago < 60. then
    "<small>now</small>"
  else if ago < 60. *. 60. then
    sprintf "<small>%.0f minutes ago</small>" (ago /. 60.)
  else if ago < 60. *. 60. *. 24. then
    sprintf "<small>%.0f hours ago</small>" (ago /. 60. /. 60.)
  else if ago < 60. *. 60. *. 24. *. 28. then
    sprintf "<small>%.0f days ago</small>" (ago /. 60. /. 60. /. 24.)
  else
    show_full_date ()

(* Mark up file sizes. *)
let markup_of_size bytes =
  sprintf "<small>%s</small>" (human_size bytes)

let clear { model = model; hash = hash } =
  model#clear ();
  Hashtbl.clear hash

let rec add ({ model = model; hash = hash } as t) name data =
  clear t;

  (* Populate the top level of the filetree.  If there are operating
   * systems from inspection, these have their own top level entries
   * followed by only unreferenced filesystems.  If we didn't get
   * anything from inspection, then at the top level we just show
   * filesystems.
   *)
  let other_filesystems =
    DeviceSet.of_list (List.map fst data.Slave.insp_all_filesystems) in
  let other_filesystems =
    List.fold_left (fun set { Slave.insp_filesystems = fses } ->
                      DeviceSet.subtract set (DeviceSet.of_array fses))
      other_filesystems data.Slave.insp_oses in

  (* Add top level operating systems. *)
  List.iter (add_top_level_os t name) data.Slave.insp_oses;

  (* Add top level left-over filesystems. *)
  DeviceSet.iter (add_top_level_vol t name) other_filesystems;

  (* Expand the first top level node. *)
  match model#get_iter_first with
  | None -> ()
  | Some row ->
      t.view#expand_row (model#get_path row)

and add_top_level_os ({ model = model; hash = hash } as t) name os =
  let markup =
    sprintf "<b>%s</b>\n<small>%s</small>\n<small>%s</small>"
      (markup_escape name) (markup_escape os.Slave.insp_hostname)
      (markup_escape os.Slave.insp_product_name) in

  let row = model#append () in
  make_node t row (Top (Slave.OS os));
  model#set ~row ~column:t.name_col markup

and add_top_level_vol ({ model = model; hash = hash } as t) name dev =
  let markup =
    sprintf "<b>%s</b>\n<small>from %s</small>"
      (markup_escape dev) (markup_escape name) in

  let row = model#append () in
  make_node t row (Top (Slave.Volume dev));
  model#set ~row ~column:t.name_col markup

(* Generic function to make an openable node to the tree. *)
and make_node ({ model = model; hash = hash } as t) row content =
  let hdata = { state=NodeNotStarted; content=content; visited=false } in
  store_hdata t row hdata;

  (* Create a placeholder "loading ..." row underneath this node so
   * the user has something to expand.
   *)
  let placeholder = model#append ~parent:row () in
  let hdata = { state=IsLeaf; content=Loading; visited=false } in
  store_hdata t placeholder hdata;
  model#set ~row:placeholder ~column:t.name_col "<i>Loading ...</i>";
  ignore (t.view#connect#row_expanded ~callback:(expand_row t))

and make_leaf ({ model = model; hash = hash } as t) row content =
  let hdata = { state=IsLeaf; content=content; visited=false } in
  store_hdata t row hdata

(* This is called when the user expands a row. *)
and expand_row ({ model = model; hash = hash } as t) row _ =
  match get_hdata t row with
  | { state=NodeNotStarted; content=Top src } as hdata ->
      (* User has opened a top level node that was not previously opened. *)

      (* Mark this row as loading, so we don't try to open it again. *)
      hdata.state <- NodeLoading;

      (* Get a stable path for this row. *)
      let path = model#get_path row in

      Slave.read_directory ~fail:(when_read_directory_fail t path)
        src "/" (when_read_directory t path)

  | { state=NodeNotStarted; content=Directory direntry } as hdata ->
      (* User has opened a filesystem directory not previously opened. *)

      (* Mark this row as loading. *)
      hdata.state <- NodeLoading;

      (* Get a stable path for this row. *)
      let path = model#get_path row in

      let src, pathname = get_pathname t row in

      Slave.read_directory ~fail:(when_read_directory_fail t path)
        src pathname (when_read_directory t path)

  | { state=(NodeLoading|IsNode) } -> ()

  (* These are not nodes so it should never be possible to open them. *)
  | { content=File _ } | { state=IsLeaf } -> assert false

  (* Node should not exist in the tree. *)
  | { state=NodeNotStarted; content=(Loading | ErrorMessage _ | Info _) } ->
      assert false

(* This is the callback when the slave has read the directory for us. *)
and when_read_directory ({ model = model } as t) path entries =
  debug "when_read_directory";

  let row = model#get_iter path in

  (* Add the entries. *)
  List.iter (
    fun direntry ->
      let { Slave.dent_name = name; dent_stat = stat; dent_link = link } =
        direntry in
      let row = model#append ~parent:row () in
      if is_directory stat.G.mode then
        make_node t row (Directory direntry)
      else
        make_leaf t row (File direntry);
      model#set ~row ~column:t.name_col (markup_of_name direntry);
      model#set ~row ~column:t.mode_col (markup_of_mode stat.G.mode);
      model#set ~row ~column:t.size_col (markup_of_size stat.G.size);
      model#set ~row ~column:t.date_col (markup_of_date stat.G.mtime);
  ) entries;

  (* Remove the placeholder "Loading" entry.  NB. Must be done AFTER
   * adding the other entries, or else Gtk will unexpand the row.
   *)
  (try
     let row = find_child_node_by_content t row Loading in
     ignore (model#remove row)
   with Invalid_argument _ | Not_found -> ()
  );

  (* The original directory entry has now been loaded, so
   * update its state.
   *)
  let hdata = get_hdata t row in
  hdata.state <- IsNode;
  set_visited t row

(* This is called instead of when_read_directory when the read directory
 * (or mount etc) failed.  Convert the "Loading" entry into the
 * error message.
 *)
and when_read_directory_fail ({ model = model } as t) path exn =
  debug "when_read_directory_fail: %s" (Printexc.to_string exn);

  match exn with
  | G.Error msg ->
      let row = model#get_iter path in
      let row = model#iter_children ~nth:0 (Some row) in

      let hdata = { state=IsLeaf; content=ErrorMessage msg; visited=false } in
      store_hdata t row hdata;

      model#set ~row ~column:t.name_col (markup_escape msg)

  | exn ->
      (* unexpected exception: re-raise it *)
      raise exn
